# coding: utf-8
import re

from tibetan_sort import TibetanSort

from ..utils.profile_entries import profile_entries
from .word_cleanup import word_cleanup

sort = TibetanSort()


def parse_corrected(in_str):
    # prepare string: replace returns and tabs and multiple spaces by a single space
    in_str = in_str.replace("\n", " ").replace("\t", " ")
    in_str = re.sub(r"\s+", " ", in_str)

    # parse
    sep_field = "/"
    parsed = []
    for token in in_str.split():
        fields = ["", "", "", "", ""]
        for num, f in enumerate(token.split(sep_field)):
            # cleanup the form and the lemma
            if (num == 0 or num == 2) and f:
                f = word_cleanup(f)
            fields[num] = f
        parsed.append(fields)
    return parsed


def extract_new_entries(in_str, profile_path):
    entries = profile_entries(profile_path)

    # parse input
    parsed = parse_corrected(in_str)

    # generate content, without duplicates
    entry_data = []
    for p in parsed:
        word = p[0]
        e_d = "\t".join(p)
        if (word not in entries or e_d not in entries[word]) and e_d not in entry_data:
            entry_data.append(e_d)

    # sort both lists
    # words = sort.sort_list(words)
    entry_data = sort.sort_list(entry_data)
    entry_data = ["# form	pos	lemma	sense	freq"] + entry_data

    return "\n".join(entry_data)



from pathlib import Path
from re import T

from bordr import rdr as r
from botok.tokenizers.wordtokenizer import WordTokenizer

from pybo.rdr.rdr_2_replace_matcher import rdr_2_replace_matcher
from pybo.hfr_cqlr_converter import cqlr2hfr

from pybo.segmentation_rule.make_rule import *


HOME = Path.home()
DIALECT_PACK_DIR = HOME / "Documents" / "pybo" / "dialect_packs"
DEFAULT_DPACK = "general"


def get_botok_segmentation(sample_text):
    """Tokenize sample text using botok tokenizer

    Args:
        sample_text (str): Input string that needs to be tokenize

    Returns:
        str: sample text with space between each tokens
    """
    wt = WordTokenizer()
    tokens = wt.tokenize(sample_text)
    segmented_sample_text = ''
    for token in tokens:
        token_with_tag = f'{token.text} '
        segmented_sample_text += token_with_tag
    segmented_sample_text = post_process_botok_segmented_data(segmented_sample_text)
    return segmented_sample_text

def post_process_botok_segmented_data(segmented_text):
    """Remove unwanted space from segmented text

    Args:
        segmented_text (str): Botok segmented text

    Returns:
        str: clean segmented text
    """
    clean_segmented_text = segmented_text.replace('\n ', '\n')
    clean_segmented_text = clean_segmented_text.replace('  ', ' ')
    return clean_segmented_text


def post_process_human_data(human_data):
    """Remove unwanted space and solves double shad(ཉིས་ཤད་) split cases

    Args:
        human_data (str): human segmented data

    Returns:
        str: clean human segmented data
    """
    human_data = human_data.replace('།  །', '།།')
    human_data = human_data.replace('  ', ' ')
    return human_data

def get_toks(seg_str):
    """Extract list of tokens from segmented string

    Args:
        seg_str (str): segmented string which can be by human or botok

    Returns:
        list: list of tokens
    """
    tokens = seg_str.split(' ')
    return tokens

def get_bilou_tag_line(human_toks, botok_toks):
    """Add bilou tags to botok tokens and join them as a string with space between each tokens

    Args:
        human_toks (list): tokens from human segmented line
        botok_toks (list): tokens from botok segmented line

    Returns:
        str: botok tokens with bilou tag separated by space
    """
    bilou_tag_line = ''
    while True:
        human_tok = human_toks[0]
        cur_tok = ''
        for tok_walker, botok_tok in enumerate(botok_toks):
            if botok_tok == human_tok:
                bilou_tag_line += f'{botok_tok}/U '
                del botok_toks[tok_walker]
                break
            elif botok_tok in human_tok:
                cur_tok += botok_tok
                if cur_tok == human_tok:
                    bilou_tag_line += f'{botok_tok}/I '
                    botok_toks = botok_toks[tok_walker+1:]
                    break
                elif tok_walker == 0:
                    bilou_tag_line += f'{botok_tok}/B '
                else:
                    bilou_tag_line += f'{botok_tok}/I '
            elif human_tok in botok_tok:
                cur_tok = ''
                bilou_tag_line += f'{botok_tok}/S '
                del botok_toks[tok_walker]
                break
            else:
                botok_toks = botok_toks[tok_walker:]
                break
        human_toks = human_toks[1:]
        if not human_toks:
            break
    return bilou_tag_line

def get_bilou_tag_data(corpus_data, human_data):
    """Corpus data get segmented by botok.
    Bilou tag is given to botok segmented data by comparing with human segmentation

    Args:
        corpus_data (str): corpus data (unsegmented data)
        human_data (str): segmented corpus data by human

    Returns:
        str: botok segmented data with bilou tag
    """
    botok_data = get_botok_segmentation(corpus_data)
    human_lines = human_data.splitlines()
    botok_lines = botok_data.splitlines()
    bilou_tag_data = ''
    for human_line, botok_line in zip(human_lines, botok_lines):
        human_toks = get_toks(human_line)
        botok_toks = get_toks(botok_line)
        bilou_tag_data += get_bilou_tag_line(human_toks, botok_toks) + '\n'
    return bilou_tag_data

def get_split_suggestions(bilou_tag_data):
    """Return all the tokens with Split tag(S)

    Args:
        bilou_tag_data (str): Botok segmented data with bilou tag

    Returns:
        list: list of tokens with split tags
    """
    split_suggestions = [split_token[:-2] for split_token in re.findall(r'\S+/S', bilou_tag_data)]
    return list(set(split_suggestions))

def get_merge_suggestions(bilou_tag_data):
    """Return all the tokens which are meant to be merge

    Args:
        bilou_tag_data (str): Botok segmented data with bilou tag

    Returns:
        list: list of tokens that are meant to be merge
    """
    merge_suggestions = [merge_suggestion for merge_suggestion,_ in re.findall(r'(\S+/B (\S+/I )+)', bilou_tag_data)]
    return list(set(merge_suggestions))

def parse_merge_suggestion(merge_suggestion):
    """Return tokens in merge suggestion

    Args:
        merge_suggestion (str): merge suggestion extracted from bilou tagged text

    Returns:
        list: tokens in merge suggestion
    """
    merge_suggestion_tokens = [token[:-2] for token in merge_suggestion.split(' ') if token]
    return merge_suggestion_tokens

def get_counter_merge_suggestion(merge_suggestion_tokens):
    """Return opposite of merge suggestion

    Args:
        merge_suggestion_tokens (list): tokens in merge suggestion

    Returns:
        str: opposite of merge suggestion
    """
    counter_merge_suggestion = ' '.join(merge_suggestion_tokens)
    if merge_suggestion_tokens[-1][-1] == '་':
        counter_merge_suggestion += " "
    return counter_merge_suggestion
    
def get_remove_word_candidates(split_suggestions, human_data):
    """Return remove word candidate or non ambiguous spilt options from spilt suggestions using human data

    Args:
        split_suggestions (list): spilt suggestion extracted from bilou tagged text
        human_data (str): human segmented text

    Returns:
        list: remove word candidates
    """
    remove_word_candidate = []
    for spilt_suggestion_token in split_suggestions:
        if spilt_suggestion_token not in human_data:
            remove_word_candidate.append(spilt_suggestion_token)
    return remove_word_candidate

def get_new_word_candidate(merge_suggestion, human_data):
    """Return new word if merge suggestion is not ambiguous one else empty string return

    Args:
        merge_suggestion (str): merge sugeestion
        human_data (str): human segmented data

    Returns:
        str: new word candidate
    """
    new_word = ''
    merge_suggestion_tokens = parse_merge_suggestion(merge_suggestion)
    counter_merge_suggestion = " " + get_counter_merge_suggestion(merge_suggestion_tokens)
    if counter_merge_suggestion not in human_data:
        new_word =  ''.join(merge_suggestion_tokens)
    return new_word

def get_new_word_candidates(merge_suggestions, human_data):
    """Return all the new word candidate from merge suggestions using human data

    Args:
        merge_suggestions (list): merge suggestions extracted from bilou tagged text
        human_data (str): human segmented data

    Returns:
        list: new word candidate
    """
    new_word_candidate = []
    for merge_suggestion in merge_suggestions:
        new_word = get_new_word_candidate(merge_suggestion, human_data)
        if new_word:
            new_word_candidate.append(new_word)     
    return new_word_candidate

def filter_seg_errors(bilou_tag_data, human_data):
    """Filters out obivious segmentation error and extract new words and new remove words

    Args:
        bilou_tag_data (str): segmented botok data with bilou tag
        human_data (ste): segmented human data

    Returns:
        list: new word list and new remove word list
    """
    new_word_candidate = []
    new_remove_word_candidate = []
    split_suggestions = get_split_suggestions(bilou_tag_data)
    merge_suggestions = get_merge_suggestions(bilou_tag_data)
    new_word_candidate = get_new_word_candidates(merge_suggestions, human_data)  
    new_remove_word_candidate = get_remove_word_candidates(split_suggestions, human_data)
    return new_word_candidate, new_remove_word_candidate

def rdr_postprocess(file_path):
    suffixes = [".DICT", ".INIT", ".RAW", ".RDR", ".sDict"]
    for s in suffixes:
        Path(file_path.parent / (file_path.name + s)).unlink()

def remove_duplicate_word(word_list):
    return list(set(word_list))

def add_word_2_adjustment(words_2_add, corpus_file_name, dialect_pack_name, type='words'):
    """New word candidates or new remove word candidates are added with existing word list.
    Duplicates are then removed.
    Unique word list are then added to its file.

    Args:
        words_2_add (list): word list of new word candidates or new remove word candidates
        corpus_file_name (str): courpus file name
        dialect_pack_name (str): current working dialect pack name
        type (str, optional): type can be either words or remove. Defaults to 'words'.
    
    Returns:
        list: latest word list of mentioned type
    """
    old_word_list = []
    word_list_path = (DIALECT_PACK_DIR / dialect_pack_name / "adjustments" / type / f'{corpus_file_name}.tsv')
    if word_list_path.is_file():
        old_word_list = [old_word for old_word in word_list_path.read_text(encoding='utf-8-sig').splitlines() if old_word]
    new_word_list = old_word_list + words_2_add
    new_word_list = remove_duplicate_word(new_word_list)
    new_words = '\n'.join(new_word_list)
    word_list_path.write_text(new_words, encoding='utf-8-sig')
    print(f'[INFO]: New {type} added to adjustment {type} list..')
    return new_word_list

def get_bilou_rules(bilou_tag_data_path):
    """Extract rdr rules by training RDR model using bilou tagged data.
    Convert rdr rules to cql rules and returning it.

    Args:
        bilou_tag_data_path (pathlib): path of bilou tagged data

    Returns:
        list: rdr rules converted into cql rules 
    """
    log = r(str(bilou_tag_data_path), mode="train", verbose=True)
    print('[INFO]: RDR TRAINING COMPLETED..')
    rdr_rules = Path(f"{bilou_tag_data_path}.RDR").read_text(
        encoding="utf-8-sig"
    )
    bilou_rules = rdr_2_replace_matcher(rdr_rules).splitlines()
    bilou_rules = list(set(bilou_rules))
    return bilou_rules

def convert_bilou_rules(bilou_rules, bilou_tag_init, human_data):
    """Convert bilou rules to normal cql rules as rules with bilou tag are not usable by botok

    Args:
        bilou_rules (list): cql rules with bilou tag
        bilou_tag_init (str): bilou tagged initial text
        human_data (str): human segmented data

    Returns:
        list: usable cql rule by botok
    """
    new_cql_rules = []
    for bilou_rule in bilou_rules:
        tokens_info, index_info, operator, conclusion = parse_rule(bilou_rule)
        tokens_in_rule = get_tokens(tokens_info)
        ambiguous_seg_candidates = get_ambiguous_seg_candidates(tokens_in_rule, index_info, bilou_tag_init)
        new_cql_rules += get_new_rule(ambiguous_seg_candidates, int(index_info)+1, conclusion, human_data)
    new_cql_rules = list(set(new_cql_rules))
    return new_cql_rules

def extract_seg_rule(corpus_file_path, dialect_pack_name=DEFAULT_DPACK, type='cql', no_epochs = 3):
    """Extracts segmentation rules.

    Args:
        corpus_file_path (pathlib): input file's path
        dialect_pack_name (string, optional): name of dialect pack for which rules are. Defaults to DEFAULT_DPACK.
        type (str, optional): type of rules can be human friendly rule(hfr) or corpus query rule. Defaults to 'cql'.
        no_epochs (int, optional): Number of times word filters need to perform. Defaults to 3.

    Returns:
        str: segmentation rules
    """
    new_word_list = []
    new_remove_word_list = []
    corpus_file_name = corpus_file_path.stem
    number_of_segmentation = 1
    human_data = (corpus_file_path.parent / f'{corpus_file_name}_hd.txt').read_text(encoding='utf-8-sig')
    human_data = post_process_human_data(human_data)
    corpus_data = corpus_file_path.read_text(encoding='utf-8-sig')
    while True:
        bilou_tag_data = get_bilou_tag_data(corpus_data, human_data)
        print(f'[INFO]: SEGMENTATION PHASE {number_of_segmentation} COMPLETED..')
        new_word_list, new_remove_word_list = filter_seg_errors(bilou_tag_data, human_data)
        print('[INFO]: FILTER SEGMENTATION ERROR COMPLETED..')
        if new_word_list:
            new_word_list = add_word_2_adjustment(new_word_list, corpus_file_name, dialect_pack_name, type='words')
        if new_remove_word_list:
            new_remove_word_list = add_word_2_adjustment(new_remove_word_list, corpus_file_name, dialect_pack_name, type='remove')
        bilou_tag_data = get_bilou_tag_data(corpus_data, human_data)
        word_list, remove_word_list = filter_seg_errors(bilou_tag_data, human_data)
        new_remove_word_list = [remove_word for remove_word in remove_word_list if remove_word not in new_remove_word_list]
        new_word_list = [word for word in word_list if word not in new_word_list]
        number_of_segmentation += 1
        if (not new_word_list and not new_remove_word_list) or number_of_segmentation > no_epochs:
            break
    bilou_tag_data_path = (corpus_file_path.parent / f'{corpus_file_name}_tr_data.txt')
    bilou_tag_data_path.write_text(bilou_tag_data, encoding='utf-8')
    bilou_rules = get_bilou_rules(bilou_tag_data_path)
    new_cql_rules = []
    bilou_tag_init = (corpus_file_path.parent / f'{bilou_tag_data_path.name}.INIT').read_text(encoding='utf-8-sig')
    new_cql_rules = convert_bilou_rules(bilou_rules, bilou_tag_init, human_data)
    new_cql_rules = "\n".join(new_cql_rules)
    rdr_postprocess(bilou_tag_data_path)
    if type != 'cql':
        new_cql_rules = cqlr2hfr(new_cql_rules)
    return new_cql_rules


import json
from pathlib import Path
from shutil import rmtree

import click
from bordr import __version__ as bordr__version
from botok import Config, Text, WordTokenizer
from botok import __version__ as botok__version__
from botok import expose_data
from pyewts import VERSION as pyewts__version__
from tibetan_sort import TibetanSort
from tibetan_sort import __version__ as tibetan_sort__version__

from pybo import __version__ as pybo__version__
from pybo.corpus.parse_corrected import extract_new_entries
from pybo.pipeline.pipes import pybo_form, pybo_mod, pybo_prep
from pybo.rdr.rdr import rdr as r
from pybo.rdr.rdr_2_replace_matcher import rdr_2_replace_matcher
from pybo.utils.profile_report import profile_report as p_report
from pybo.utils.regex_batch_apply import batch_apply_regex, get_regex_pairs
from pybo.hfr_cqlr_converter import cqlr2hfr, hfr2cqlr
from pybo.segmentation_rule.pipeline import extract_seg_rule

HOME = Path.home()
DIALECT_PACK_DIR = HOME / "Documents" / "pybo" / "dialect_packs"
DEFAULT_DPACK = "general"
CONFIG_DIR = HOME / ".pybo"
CONFIG_FILE = CONFIG_DIR / "config.json"


@click.group()
@click.version_option(pybo__version__)
def cli():
    pass


@cli.command()
def info():
    click.echo("pybo install path: " + str(Path(__file__).parent.resolve()))
    click.echo("pybo: " + pybo__version__)
    click.echo("botok: " + botok__version__)
    click.echo("pyewts: " + pyewts__version__)
    click.echo("bordr: " + bordr__version)
    click.echo("tibetan_sort: " + tibetan_sort__version__)


def prepare_folder(main=None, custom=None, overwrite=False):
    profile = "POS"
    # 1. MAIN PROFILE
    if not main:
        # for better Windows support:
        # https://stackoverflow.com/questions/6227590/finding-the-users-my-documents-path/6227623#6227623
        main = Path.home() / "Documents/pybo/main"
    else:
        main = Path(main)
    main.mkdir(parents=True, exist_ok=True)

    if overwrite:
        rmtree(main, ignore_errors=True)
        main.mkdir()

    try:
        expose_data(main, profile=profile)
    except IOError:
        click.echo('using the existing data in "Documents/pybo/main/"')

    # 2. CUSTOM PROFILE
    if not custom:
        custom = Path.home() / "Documents/pybo/custom"
    else:
        custom = Path(custom)
    custom.mkdir(exist_ok=True)
    for dir in ["adjustment", "remove", "words", "words_skrt"]:
        Path(custom / dir).mkdir(exist_ok=True)

    return main, custom


def save_config(dialect_pack_path):
    config = {"dialect_pack_path": str(dialect_pack_path)}
    if not CONFIG_DIR.is_dir():
        CONFIG_DIR.mkdir(parents=True)
    json.dump(config, CONFIG_FILE.open("w"))


def load_config():
    if not CONFIG_FILE.is_file():
        return
    else:
        config = json.load(CONFIG_FILE.open())
    return config


# Tokenize file
@cli.command()
@click.argument("input-dir", type=click.Path(exists=True))
@click.option(
    "-t",
    "--tags",
    help="""Select and order the tags. Available tags are:
t-clean_text, p-pos, l-lemma, s-sense.\n
Usage: `-t tpl` will give for every token `<raw-text>/<clean-text>/<pos>/<lemma>`
and will give just `<raw-text>` if tag option is not specified.""",
)
@click.option(
    "-o", type=click.Path(exists=True), help="output dir, default is the input_dir"
)
@click.option("-d", "--dialect-name", type=str, help="official dialect pack name.")
@click.option(
    "-p",
    "--dialect-path",
    type=click.Path(exists=True),
    help="path to the dialect pack",
)
@click.option("-w", "--overwrite", is_flag=True)
@click.option("-r", "--rebuild-trie", is_flag=True)
def tok(**kwargs):
    input_dir = Path(kwargs["input_dir"])
    dialect_name = kwargs["dialect_name"]
    dialect_path = kwargs["dialect_path"]
    # overwrite = kwargs["overwrite"]
    rebuild = kwargs["rebuild_trie"]

    # load botok config
    if dialect_name:
        config = Config(dialect_name=dialect_name)
        save_config(config.dialect_pack_path)
    elif dialect_path:
        config = Config.from_path(dialect_path)
        # config.dialect_pack_path = Path(dialect_pack_path)
        save_config(config.dialect_pack_path)
    else:
        pybo_config = load_config()
        if not pybo_config:
            config = Config()
            save_config(config.dialect_pack_path)
        else:
            dialect_pack_path = pybo_config["dialect_pack_path"]
            config = Config.from_path(dialect_pack_path)

    print(
        f"[INFO] Using `{config.dialect_pack_path.name}` dialect pack for tokenization ..."
    )

    wt = WordTokenizer(config=config, build_trie=rebuild)

    def pybo_tok(in_str):
        return wt.tokenize(in_str)

    # Select and Order the tags
    if kwargs["tags"]:
        pybo_mod.__defaults__ = (list(kwargs["tags"]),)

    if input_dir.is_dir():
        if kwargs["o"] is not None:
            output_dir = Path(kwargs["o"])
        else:
            output_dir = input_dir.parent / (input_dir.name + "_tok")
            output_dir.mkdir(exist_ok=True)
        for f in input_dir.glob("*.txt"):
            out_file = output_dir / (f.stem + "_tok.txt")
            text = Text(f, out_file)
            text.custom_pipeline(pybo_prep, pybo_tok, pybo_mod, pybo_form)
    elif input_dir.is_file():
        input_file = input_dir
        if kwargs["o"] is not None:
            output_dir = Path(kwargs["o"])
        else:
            output_dir = input_file.parent / (input_file.stem + "_tok")
            output_dir.mkdir(exist_ok=True)
        out_file = output_dir / (input_file.stem + "_tok.txt")
        text = Text(input_file, out_file)
        text.custom_pipeline(pybo_prep, pybo_tok, pybo_mod, pybo_form)
    else:
        print("[INFO] Invalid input directory or file!!!")


# Tokenize string
@cli.command()
@click.argument("string")
def tok_string(**kwargs):
    t = Text(kwargs["string"])
    click.echo(t.tokenize_words_raw_lines)


# lists
tag_types = ["pos", "lemma", "sense"]


@cli.command()
@click.argument("input-dir", type=click.Path(exists=True))
@click.option("-t", "--type")
def lists(**kwargs):
    path = Path(kwargs["path"])

    text_string = ""
    for f in path.glob("*.txt"):
        text_string += f.read_text(encoding="utf-8-sig")


# create report for botok profiles
@cli.command()
@click.argument("profile", type=click.Path(exists=True))
def profile_report(**kwargs):
    p_report(kwargs["profile"])


# rdr_2_replace_matcher
@cli.command()
@click.argument("infile", type=click.Path(exists=True))
def rdr2repl(**kwargs):
    infile = Path(kwargs["infile"])
    outfile = infile.parent / (infile.stem + ".yaml")
    dump = infile.read_text(encoding="utf-8-sig")
    processed = rdr_2_replace_matcher(dump)
    outfile.write_text(processed, encoding="utf-8-sig")


# sort in the Tibetan order
@cli.command()
@click.argument("infile", type=click.Path(exists=True))
def kakha(**kwargs):
    sort = TibetanSort()
    infile = Path(kwargs["infile"])
    words = infile.read_text(encoding="utf-8-sig").split()
    print(f"Sorting {infile.name}")
    words = sort.sort_list(words)
    print(f"{infile.name} is sorted")
    infile.write_text("\n".join(words), encoding="utf-8-sig")


# generate rdr rules
@cli.command()
@click.argument("input", type=click.Path(exists=True))
@click.option("-dp", type=str, help="Dialect pack name, default is general")
@click.option("-k", "--keep", type=str)
@click.option('--type', type=str, help="Type can be either cql which is default type or hfr(Human friendly rule)")
def extract_rules(**kwargs):
    file_or_dir = Path(kwargs["input"])
    dialect_pack_name = kwargs["dp"] if kwargs["dp"] else DEFAULT_DPACK
    keep = "none" if kwargs["keep"] is None else kwargs["keep"]
    type = "cql" if kwargs["type"] is None else kwargs["type"]
    if type == "cql":
        out_dir = DIALECT_PACK_DIR / dialect_pack_name / "adjustments" / "rules"
    else:
        out_dir = DIALECT_PACK_DIR / dialect_pack_name / "hfr_rules"
        out_dir.mkdir(exist_ok=True)

    log = None
    click.echo("[INFO] Extracing adjustments rules ...")
    if file_or_dir.is_dir():
        file = file_or_dir / file_or_dir.name
        with open(file, encoding="utf-8-sig", mode="w") as tmp:
            for f in file_or_dir.glob("*.txt"):
                tmp.write(f.read_text(encoding="utf-8-sig") + " ")
        log = r(file, outdir=out_dir, keep=keep, type=type)
        file.unlink()
    elif file_or_dir.is_file():
        log = r(file_or_dir, out_dir, keep=keep, type=type)
        click.echo(f"[INFO] {file_or_dir} does not exist!")

    click.echo(log)
    click.echo("[INFO] Completed !")
    click.echo(f"[INFO] Added adjustments rules to {dialect_pack_name}")

# generate rdr rules
@cli.command()
@click.argument("input", type=click.Path(exists=True))
@click.option("-dp", type=str, help="Dialect pack name, default is general")
@click.option('--type', type=str, help="Type can be either cql which is default type or hfr(Human friendly rule)")
@click.option("--e", type=int)
def extract_seg_rules(**kwargs):
    rules = ''
    input_path = Path(kwargs["input"])
    dialect_pack_name = kwargs["dp"] if kwargs["dp"] else DEFAULT_DPACK
    type = "cql" if kwargs["type"] is None else kwargs["type"]
    epochs = 3 if kwargs['e'] is None else kwargs['e']
    if type == "cql":
        out_dir = DIALECT_PACK_DIR / dialect_pack_name / "adjustments" / "rules"
    else:
        out_dir = DIALECT_PACK_DIR / dialect_pack_name / "hfr_rules"
        out_dir.mkdir(exist_ok=True)

    click.echo("[INFO] Extracing adjustments rules ...")
    
    if input_path.is_dir():
        print('[ERROR] Invalid file name!!')
    elif input_path.is_file():
        rules += extract_seg_rule(input_path, dialect_pack_name, type, epochs)
        if rules:
            (out_dir / f'{input_path.stem}_rules.tsv').write_text(rules, encoding='utf-8')
        else:
            print('[INFO] No rules found')
        
    click.echo("[INFO] Completed !")
    click.echo(f"[INFO] Added adjustments rules to {dialect_pack_name}")

#convert cql to hfr
@cli.command()
@click.argument("input", type=click.Path(exists=True))
@click.option("-dp", type=str, help="Dialect pack name, default is general")
def convert_cql2hfr(**kwargs):
    cql_path = Path(kwargs['input'])
    dialect_pack_name = kwargs["dp"] if kwargs["dp"] else DEFAULT_DPACK
    hfr_dir = DIALECT_PACK_DIR / dialect_pack_name / "hfr_rules"
    hfr_dir.mkdir(exist_ok=True) 
    hfr_file_path = hfr_dir / (cql_path.stem + ".tsv")
    cql_rules = cql_path.read_text(encoding='utf-8')
    hfr = cqlr2hfr(cql_rules)
    hfr_file_path.write_text(hfr, encoding='utf-8')

#convert hfr to cql
@cli.command()
@click.argument("input", type=click.Path(exists=True))
@click.option("-dp", type=str, help="Dialect pack name, default is general")
def convert_hfr2cql(**kwargs):
    hfr_path = Path(kwargs['input'])
    dialect_pack_name = kwargs["dp"] if kwargs["dp"] else DEFAULT_DPACK
    cql_dir = DIALECT_PACK_DIR / dialect_pack_name / "adjustments" / "rules"
    cql_dir.mkdir(exist_ok = True)
    cql_file_path = cql_dir / (hfr_path.stem + ".tsv")
    hfr = hfr_path.read_text(encoding='utf-8')
    cql = hfr2cqlr(hfr)
    cql_file_path.write_text(cql, encoding='utf-8')


# extract new entries from manually corrected texts + existing profile
@cli.command()
@click.argument("corrected-path", type=click.Path(exists=True))
@click.argument("dialect_path", type=click.Path(exists=True))
@click.option("-o", "--out-dir", type=click.Path(exists=True))
def profile_update(**kwargs):
    corrected = Path(kwargs["corrected_path"])
    dialect_path = Path(kwargs["dialect_path"])
    out_dir = Path(kwargs["out_dir"]) if kwargs["out_dir"] else None

    dump = ""
    for f in corrected.glob("*.txt"):
        dump += f.read_text(encoding="utf-8-sig") + "\n"

    rules = extract_new_entries(dump, dialect_path)
    if not out_dir:
        out = corrected.parent / (corrected.name + "_words.tsv")
    else:
        out = out_dir / (corrected.name + "_words.tsv")

    if not out.parent.is_dir():
        out.parent.mkdir(exist_ok=True)

    out.write_text(rules, encoding="utf-8-sig")


# FNR - Find and Replace with a list of regexes
@cli.command()
@click.argument("in-dir", type=click.Path(exists=True))
@click.argument("regex-file", type=click.Path(exists=True))
@click.option("-o", "--out-dir", type=click.Path(exists=True))
@click.option("-t", "--tag")
def fnr(**kwargs):
    # get the args
    indir = Path(kwargs["in_dir"])
    regex_file = Path(kwargs["regex_file"])
    out_dir = Path(kwargs["out_dir"]) if kwargs["out_dir"] else None

    if not indir.is_dir():
        click.echo("IN_DIR should be a folder, not a file.\nexiting...")
        exit(1)

    # optional out file tag
    tag = kwargs["tag"] if kwargs["tag"] else regex_file.stem

    # generate rules
    rules = get_regex_pairs(regex_file.open(encoding="utf-8-sig").readlines())

    # apply on each file, prefixing each one with the regex filename
    for f in indir.rglob("*.txt"):
        if not f.stem.startswith("_"):
            string = f.read_text(encoding="utf-8-sig")
            out = batch_apply_regex(string, rules)
            name = f"_{tag}_" + f.name
            if out_dir:
                Path(out_dir).mkdir(parents=True, exist_ok=True)
                outfile = out_dir / name
            else:
                outfile = f.parent / name
            outfile.write_text(out, encoding="utf-8-sig")


if __name__ == "__main__":
    # cli()
    save_config("test_path")



# coding: utf-8
from collections import defaultdict
from pathlib import Path


def reorder_data(data):
    ordered = []
    for entry, e in data.items():
        count = 0
        for _, files in e.items():
            count += len(files)
        ordered.append((count, {entry: e}))
    ordered = sorted(ordered, reverse=True, key=lambda x: x[0])
    return ordered


def profile_report(pathname):
    pathname = Path(pathname)
    data = {}

    for d in sorted(pathname.glob("*")):
        # filter unwanted directories and files
        dirs_ignored = ["adjustment", "entry_data"]
        if not d.is_dir() or d.name in dirs_ignored or d.name.startswith("."):
            continue

        for f in sorted(d.glob("*.tsv")):
            lines = f.read_text(encoding="utf-8-sig").splitlines()
            for num, line in enumerate(lines):
                if line.startswith("#"):
                    continue
                entry = line.split("\t", 1)[0]
                path = f"{d.name}/{f.name}"

                if entry not in data:
                    data[entry] = {}
                if line not in data[entry]:
                    data[entry][line] = []

                data[entry][line].append((path, num))

    data = reorder_data(data)

    # filter and format all entries that have similar forms over files
    report = ["WORD\tENTRY\tFILE-NAME\tLINE-NUMBER"]
    count = defaultdict(int)
    for num, d in data:
        count[num] += 1
        for entry, e in d.items():
            tmp = []
            tmp.append(f"{entry}: {num}")
            for line, files in e.items():
                tmp.append(f'\t"{line}"')
                tmp.extend([f"\t\t{f}\t{n}" for f, n in files])
            report.extend(tmp)
    report = (
        [f"total distinct entries: {len(data)}"]
        + [f"entries with {a} entries: {b}" for a, b in count.items()]
        + [""]
        + report
    )
    report = "\n".join(report)

    # print to file
    out = pathname / (pathname.name + "_report.tsv")
    out.write_text(report, encoding="utf-8-sig")



import re
from pathlib import Path

cql2hfr_tag = {
    '"ADJ"': "རྒྱན",
    '"ADP"': "སྦྱོར",
    '"ADV"': "བསྣན",
    '"AUX"': "གྲོགས",
    '"CCONJ"': "སྦྲེལ",
    '"DET"': "ངེས",
    '"INTJ"': "འབོད",
    '"NOUN"': "མིང",
    '"NUM"': "གྲངས",
    '"PRON"': "ཚབ",
    '"PROPN"': "ཁྱད",
    '"PUNCT"': "ཚེག",
    '"SCONJ"': "ལྟོས",
    '"VERB"': "བྱ",
    '"PART"': "རོགས",
    "pos=": "གཤིས=",
    "lemma=": "མ=",
    "sense=": "དོན=",
    "&": "༈",
    "[": "༺",
    "]": "༻",
}


def cqlr2hfr(cqlr_string):
    """Convert corpus queery language(cql) rules to human friendly rules(hfr) which has UDPOS in Tibetan.

    Args:
        cql_string (str): corpus queery language rules

    Returns:
        str: human friendly rules(in Tibetan language)
    """
    hfr_string = cqlr_string
    for cql_tag, hfr_tag in cql2hfr_tag.items():
        hfr_string = hfr_string.replace(cql_tag, hfr_tag)
    return hfr_string


def hfr2cqlr(hfr_string):
    """Convert human friendly rules(hfr) to corpus queery language rules format.

    Args:
        hfr_string (str): Human friendly rules(hfr)

    Returns:
        str: Corpus queery language(cql) rules format.
    """
    cql_string = hfr_string
    for cql_tag, hfr_tag in cql2hfr_tag.items():
        cql_string = cql_string.replace(hfr_tag, cql_tag)
    return cql_string



import csv
import re

from botok import Text

ID = -1

POS_NAMES = (" མིང་ཚིག ", " བྱ་ཚིག ", " བྱེད་ཚིག ", " གྲོགས་ཚིག ")


class cols:
    ID = "ID"
    FORM = "Form"
    LEMMA = "Lemma"
    MON_POS = "MonPOS"
    MON_FEATURE = "MonFeature"
    MON_TAG = "MonTag"
    POS = "POS"
    FEATURE = "Feature"
    MORPH = "Morph"
    SENSE_TAG = "SenseTag"
    DEFINITION = "Definition"
    EXAMPLE = "Example"


def create_word(
    form,
    lemma,
    mon_pos=None,
    mon_feature=None,
    mon_tag=None,
    pos=None,
    feature=None,
    morph=None,
    sense_tag=None,
    definition=None,
    example=None,
):
    global ID
    ID += 1
    return {
        "ID": ID,
        "Form": form,
        "Lemma": lemma,
        "MonPOS": mon_pos,
        "MonFeature": mon_feature,
        "MonTag": mon_tag,
        "POS": pos,
        "Feature": feature,
        "Morph": feature,
        "SenseTag": sense_tag,
        "Definition": definition,
        "Example": example,
    }


def csv_loader(path):
    with open(path, "r") as csv_file:
        reader = csv.reader(csv_file)
        for i, row in enumerate(reader):
            if i == 0:
                continue
            yield row


def get_single_pos(chunk_containing_pos):
    """Return only first pos and it's content."""
    pos_char_end_idx = chunk_containing_pos.find(" ")
    pos = chunk_containing_pos[:pos_char_end_idx]
    pos_content = chunk_containing_pos[(pos_char_end_idx + 1) :]
    return pos, pos_content


def find_all_remaining_pos(chunk):
    """Return all pos position and it's length.

    Return:
        pos_start_idxs (list): [(pos_start_idx, len(pos_name)), ...] sorted on pos_start_idx.

    """
    pos_start_idxs = []
    pos_found = False
    for pos_name in POS_NAMES:
        pos_found = True
        pos_start_idx = chunk.find(pos_name)
        if pos_start_idx != -1:
            pos_start_idxs.append((pos_start_idx, len(pos_name)))
    if pos_found:
        pos_start_idxs.append((len(chunk), 0))
    return sorted(pos_start_idxs, key=lambda x: x[0])


def get_pos_list(text):
    """Parse pos and it's content (mon_tags, definitions) in string.

    Returns:
        post_list (list): [(pos, pos_content), ...]
    """

    pos_list = []
    chunks_containing_pos = text.split(" 1. ")
    estimated_n_pos = len(chunks_containing_pos)
    if estimated_n_pos == 1:  # one_pos_one_sense
        chunk_containing_pos = chunks_containing_pos[0].strip()
        pos, pos_content = get_single_pos(chunk_containing_pos)
        pos_list.append((pos, pos_content))
    elif estimated_n_pos == 2:  # one_pos_multi_senses
        pos, pos_content = chunks_containing_pos
        pos_list.append((pos, pos_content))
    else:  # multi_pos_multi_senses
        pos = chunks_containing_pos[0]
        for i, chunk_containing_pos in enumerate(chunks_containing_pos[1:]):
            if i == estimated_n_pos - 2:  # if last chunk, check for all pos
                new_chunk_start = 0
                next_pos_start_idxs = find_all_remaining_pos(chunk_containing_pos)
                for next_pos_start_idx, pos_name_len in next_pos_start_idxs:
                    pos_content = chunk_containing_pos[
                        new_chunk_start:next_pos_start_idx
                    ]
                    pos_list.append((pos, pos_content))
                    pos = chunk_containing_pos[
                        next_pos_start_idx : next_pos_start_idx + pos_name_len
                    ].strip()
                    new_chunk_start = next_pos_start_idx + pos_name_len
                if not next_pos_start_idxs:
                    pos_list.append((pos, chunk_containing_pos))
            else:
                next_pos_start_idx = chunk_containing_pos.rfind(" ")
                pos_content = chunk_containing_pos[:next_pos_start_idx]
                pos_list.append((pos, pos_content))
                pos = chunk_containing_pos[next_pos_start_idx + 1 :]

    return pos_list


def get_definition_list(pos_list):
    """Parse definitions from pos_content.

    Returns:
        definition_list (list): [(pos, definition-content), ...]

    """
    definition_list = []
    for pos, pos_content in pos_list:
        for definition_content in re.split(r" \d\. ", pos_content):
            definition_list.append((pos, definition_content))
    return definition_list


def get_tag_list(definition_list):
    """Parse monlam tag from definition content.

    Returns:
        tag_list (list): [(pos, tag, definition), ...]
    """

    def parse_tag(text):
        if text[0] != "༡":
            return "", text
        tag_end_idx = text.find(" ")
        tag = text[:tag_end_idx]
        definition = text[tag_end_idx + 1 :]
        return tag, definition

    tag_list = []
    for pos, definition_content in definition_list:
        tag, definition = parse_tag(definition_content)
        tag_list.append((pos, tag, definition))
    return tag_list


def get_sense_tag_list(tag_list):
    """Parse sense from definition.

    Sense here the first word of the given definition.

    Returns:
        sense_tag_list (list): [(pos, tag, sense_tag, definition), ...]

    """

    def get_first_segment(text, delimiter=" "):
        seg_idx = text.find(delimiter)
        if seg_idx == -1:
            return text
        return text[:seg_idx]

    sense_tag_list = []
    for *pos_and_tag, definition in tag_list:
        first_segment = get_first_segment(definition)
        tokenized_segment = Text(first_segment).tokenize_words_raw_text
        sense = get_first_segment(tokenized_segment)
        sense_tag_list.append((*pos_and_tag, sense, definition))
    return sense_tag_list


def get_example_list(sense_tag_list):
    """Parse example from the definition."""

    def parse_example(text, example_tag="དཔེར་ན།"):
        example_start_idx = text.rfind(example_tag)
        if example_start_idx == -1:
            return text, ""
        definition = text[:example_start_idx].strip()
        example = text[example_start_idx + len(example_tag) :].strip()
        return definition, example

    example_list = []
    for *pos_tag_sense, definition in sense_tag_list:
        definition, example = parse_example(definition)
        example_list.append((*pos_tag_sense, definition, example))
    return example_list


def parse_attrs(form, text_containing_attrs):
    pos_list = get_pos_list(text_containing_attrs)
    definition_list = get_definition_list(pos_list)
    tag_list = get_tag_list(definition_list)
    sense_tag_list = get_sense_tag_list(tag_list)
    example_list = get_example_list(sense_tag_list)
    return example_list


def monlam2wordlist(rows):
    word_list_rows = []
    for row in rows:
        *_, form, result = row
        attrs = parse_attrs(form, result)
        print(row, attrs)
    return word_list_rows


def dump_tsv(rows, out_path):
    with open(out_path, "w") as csv_file:
        writer = csv.writer(csv_file, delimiter="\t")
        writer.writerows(rows)



# coding: utf-8

import pyewts
from botok import *

from .corpus.parse_corrected import extract_new_entries, parse_corrected
from .pipeline.pipes import pybo_form, pybo_mod, pybo_prep
from .utils.profile_report import profile_report
from .utils.regex_batch_apply import batch_apply_regex, get_regex_pairs

__version__ = "0.7.13"



#! /usr/bin/env python
# -*- coding: utf8 -*-

from __future__ import print_function

import re
from pathlib import Path

import setuptools
from pkg_resources import parse_version

assert parse_version(setuptools.__version__) >= parse_version("38.6.0")


def get_version(prop, project):
    project = Path(__file__).parent / project / "__init__.py"
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), project.read_text()
    )
    return result.group(1)


def read(fname):
    p = Path(__file__).parent / fname
    with p.open(encoding="utf-8") as f:
        return f.read()


setuptools.setup(
    name="pybo",
    version=get_version("__version__", "pybo"),  # edit version in pybo/__init__.py
    author="Esukhia development team",
    author_email="esukhiadev@gmail.com",
    description="Python utils for processing Tibetan",
    license="Apache2",
    keywords="nlp computational_linguistics search ngrams language_models linguistics toolkit tibetan",
    url="https://github.com/Esukhia/pybo",
    packages=setuptools.find_packages(),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    project_urls={
        "Source": "https://github.com/Esukhia/pybo",
        "Tracker": "https://github.com/Esukhia/pybo/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Text Processing :: Linguistic",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: Tibetan",
    ],
    python_requires=">=3.6",
    install_requires=["botok>=0.8.2", "pyyaml", "click", "pyewts", "bordr", "tibetan_sort", "pytest"],
    tests_require=["pytest"],
    entry_points={
        "console_scripts": ["bo=pybo.cli:cli"]  # command=package.module:function
    },
)



import re

def parse_rule(rule):
    """Parse all the components of cql rule

    Args:
        rule (str): cql rule

    Returns:
        str: token info, index info, operator of the rule and conclusion tag
    """
    part_of_rules = rule.split('\t')
    return part_of_rules[0], part_of_rules[1], part_of_rules[2], part_of_rules[3]

def get_tokens(tokens_info):
    """Parse tokens from tokens info of a cql rule

    Args:
        tokens_info (str): tokens info in a cql rule

    Returns:
        list: tokens from token info
    """
    tokens = re.findall(r'\[.*?\]', tokens_info)
    return tokens

def parse_tok(token):
    try:
        bilou_tag = re.search(r'pos="(\S)"', token).group(1)
    except:
        bilou_tag = ''
    try:
        text = re.search(r'text="(\S+)" ?', token).group(1)
    except:
        text = ''
    return bilou_tag, text

def add_extra_token_pat(ambiguous_seg_pat, index_info):
    extra_token_pat = r' \S+?/\S '
    ambiguous_seg_pat_with_extra_token_pat = f'{extra_token_pat}{ambiguous_seg_pat}{extra_token_pat}'
    ambiguous_seg_pat_with_extra_token_pat = ambiguous_seg_pat_with_extra_token_pat.replace('  ', ' ')
    return ambiguous_seg_pat_with_extra_token_pat

def get_ambiguous_seg_pat(tokens_in_rule, index_info):
    """Return ambguous segmentation's pattern

    Args:
        tokens_in_rule (list): tokens in bilou rule

    Returns:
        str: ambiguos segmentation's pattern
    """
    ambiguous_seg_pat = ''
    for token in tokens_in_rule:
        bilou_tag, text = parse_tok(token)
        if text:
            ambiguous_seg_pat += f' {text}'
            if bilou_tag:
                ambiguous_seg_pat += f'/{bilou_tag}'
            else:
                ambiguous_seg_pat += r'/\S'
        else:
            ambiguous_seg_pat += r" \S+?"
            if bilou_tag:
                ambiguous_seg_pat += f'/{bilou_tag}'
            else:
                ambiguous_seg_pat += r'/\S'
    if len(tokens_in_rule) < 5:
        ambiguous_seg_pat = add_extra_token_pat(ambiguous_seg_pat, index_info)
    return ambiguous_seg_pat

def construct_token_info(ambiguous_seg_candidate):
    """Construct token info part of a cql rule

    Args:
        ambiguous_seg_candidate (list): ambiguous segmentation candidate's token list

    Returns:
        str: token info part of a cql rule
    """
    token_info = ''
    for token in ambiguous_seg_candidate:
        token_parts = token.split('/')
        token_info += f'["{token_parts[0].replace(" ", "")}"] '
    return token_info.strip()

def get_ambiguous_seg_candidates(tokens_in_rule, index_info, bilou_tag_data):
    """Return all the possible ambiguous segmentation candidates containing tokens in rule

    Args:
        tokens_in_rule (list): tokens in bilou rule
        bilou_tag_data (str): bilou tagged data

    Returns:
        list: ambiguous segmentation candidates
    """
    ambiguous_seg_candidates_tokens = []
    ambiguous_seg_pat = get_ambiguous_seg_pat(tokens_in_rule, index_info)
    ambiguous_seg_candidates = re.findall(ambiguous_seg_pat, bilou_tag_data)
    ambiguous_seg_candidates = list(set(ambiguous_seg_candidates))
    for ambiguous_seg_candidate in ambiguous_seg_candidates:
        ambiguous_seg_candidates_tokens.append([token for token in ambiguous_seg_candidate.split(' ') if token])
    return ambiguous_seg_candidates_tokens

def is_single_syl(token):
    """Check token is single syllable

    Args:
        token (str): token

    Returns:
        boolean: True if token is single syllable else False
    """
    syls = [syl for syl in token.split('་') if syl]
    if len(syls) > 1:
        return False
    else:
        return True

def parse_index_info(index_info):
    """Return index of the token from index info

    Args:
        index_info (str): index info of a cql rule

    Returns:
        int: index of token
    """
    if '-' in index_info:
        index_info_parts = index_info.split('-')
        index = int(index_info_parts[0])
    else:
        index = int(index_info)
    return index

def get_counter_split_suggestion(spilt_suggestion):
    """Return opposite of split suggestion

    Args:
        spilt_suggestion (str): split suggestion

    Returns:
        str: opposite of split suggestion
    """
    syls = [syl.strip() for syl in spilt_suggestion.split('་') if syl and syl != ' ']
    suggestion = f'{syls[0]}་ {"་".join(syls[1:])}'
    if spilt_suggestion[-2] == '་':
        suggestion += '་'
    counter_split_suggestion = f' {suggestion} '
    return counter_split_suggestion

def is_invalid_split(tokens_info, index_info, human_data):
    """Return false if split suggestion is ambiguous segmentation else true 

    Args:
        tokens_info (str): token info of a rule
        index_info (str): index info of a cql rule
        human_data (str): human segmented data

    Returns:
        boolean: True if invalid split rule else False
    """
    index = parse_index_info(index_info)
    tokens = get_tokens(tokens_info)
    token_to_split = re.search(r'\"(\S+)\"', tokens[index-1]).group(1)
    if is_single_syl(token_to_split) or len(tokens) < index:
        return True
    else:
        split_suggestion = f" {token_to_split} "
        counter_split_suggestion = get_counter_split_suggestion(split_suggestion)
        if split_suggestion in human_data and counter_split_suggestion in human_data:
            return False
        else:
            return True

def is_invalid_merge(tokens_info, index_info, human_data):
    """Return false if merge suggestion is ambiguous segmentation else true 

    Args:
        tokens_info (str): token info of a rule
        index_info (str): index info of a cql rule
        human_data (str): human segmented data

    Returns:
        boolean: True if invalid merge rule else False
    """
    index = parse_index_info(index_info)
    tokens = get_tokens(tokens_info)
    if len(tokens) <= index or index == 0:
        return True
    else:
        part1 = re.search(r'\"(\S+)\"', tokens[index-1]).group(1)
        part2 = re.search(r'\"(\S+)\"', tokens[index]).group(1)
        merge_suggestion = f' {part1}{part2} '
        counter_merge_suggestion = f' {part1} {part2} '
        if "།" not in merge_suggestion and (merge_suggestion in human_data and counter_merge_suggestion in human_data):
            return False
        else:
            return True

def filter_valid_rules(new_rules, human_data):
    """Return valid rules which can solve ambiguous segmentation errors

    Args:
        new_rules (list): cql rules
        human_data (str): human segmented data

    Returns:
        list: cql rules
    """
    valid_rules = []
    for new_rule in new_rules:
        tokens_info, index_info, operator, conclusion = parse_rule(new_rule)
        if ":" == operator:
            if not is_invalid_split(tokens_info, index_info, human_data):
                valid_rules.append(new_rule)
        elif "+" == operator:
            if not is_invalid_merge(tokens_info, index_info, human_data):
                valid_rules.append(new_rule)
    return valid_rules

def get_new_rule(ambiguous_seg_candidates, index, conclusion, human_data):
    """Return list of usable cql rules by botok

    Args:
        ambiguous_seg_candidates (list): ambiguous segmentation candidates
        index (int): index of token on which operation needs to perform
        conclusion (str): conclusion tag of rule
        human_data (str): human segmented data

    Returns:
        list: usable cql rules of botok
    """
    new_rules = []
    for ambiguous_seg_candidate in ambiguous_seg_candidates:
        new_rule = f"{construct_token_info(ambiguous_seg_candidate)}\t"
        if 'B' in conclusion:
            new_rule += f'{index}\t+\t[]'
        elif 'I' in conclusion:
            new_rule += f'{index-1}\t+\t[]'
        elif 'S' in conclusion:
            new_rule += f'{index}-1\t:\t[] []'
        else:
            new_rule = ''
        if new_rule:
            new_rules.append(new_rule)
    unique_rules = list(set(new_rules))
    filtered_rules = filter_valid_rules(unique_rules, human_data)
    return filtered_rules


# coding: utf-8
import botok


def get_chunks(raw_string):
    chunker = botok.Chunks(raw_string)
    chunks = chunker.make_chunks()
    chunks = chunker.get_readable(chunks)
    return chunks


def shelve_info(chunks):
    shelved = []
    clean_chunks = []

    syl_count = 0
    for i, chunk in enumerate(chunks):
        marker, text = chunk
        if marker == "TEXT" or marker == "PUNCT":
            syl_count += 1

        # 2.a. extract transparent chars
        # TODO: adapt to also include \t as transparent char
        if "\n" in text:
            # remove transparent char
            text = text.replace("\n", "")
            index = (syl_count, "\n")

            shelved.append(index)
            clean_chunks.append((marker, text))

        # 2.b. extract any non-bo chunk
        elif marker != "TEXT" and marker != "PUNCT":
            index = (syl_count, text)
            shelved.append(index)

        else:
            clean_chunks.append(chunk)

    return clean_chunks, shelved


def pybo_prep(in_str):
    # 1. get chunks
    chunks = get_chunks(in_str)

    # 2. shelve needed info
    chunks, shelved = shelve_info(chunks)
    pybo_form_sep = pybo_form.__defaults__[0]
    pybo_form.__defaults__ = (pybo_form_sep, shelved)

    # 3. tokenize
    str_for_botok = "".join([c[1] for c in chunks])

    return str_for_botok


def get_tag(token, tag_code):
    maps = {"r": "text", "t": "text_cleaned", "p": "pos", "l": "lemma", "s": "sense"}
    try:
        return token[maps[tag_code]]
    except Exception:
        return ""


def pybo_mod(tokens, tag_codes=[]):
    """extract text/pos tuples from Token objects"""
    txt_tags = []
    for token in tokens:
        tags = []
        tags.append(token.text)
        # Select and order the tags
        for tag_code in tag_codes:
            tags.append(get_tag(token, tag_code))
        txt_tags.append(tags)
    return txt_tags


def ws2uc(tags):
    """Convert whitespace in raw-text to underscore."""
    tags[0] = tags[0].replace(" ", "_")
    return tags


def n_chunks(token):
    return len([chunk for chunk in token.split("་") if chunk])


def pybo_form(tokens, sep=" ", shelved=None):
    """Format in a single string to be written to file"""
    if not shelved:
        print(shelved)
        out = []
        shelved_idx = 0
        syl_count = 0

        # reinsert shelved tokens
        for token in tokens:
            out.append("/".join(ws2uc(token)))
            syl_count += n_chunks(token[0])
            sheveled_syl_count, shelved_cleaned_chunk = shelved[shelved_idx]
            if "PART" not in token and sheveled_syl_count <= syl_count:
                out.append(ws2uc([shelved_cleaned_chunk])[0])
                shelved_idx += 1

        # add all the remaining sheveld tokens
        if shelved_idx < len(shelved):
            for _, shelved_cleaned_chunk in shelved_cleaned_chunk[shelved_idx:]:
                out.append(ws2uc([shelved_cleaned_chunk]))
    else:
        out = ["/".join(ws2uc(token)) for token in tokens]
    return sep.join(out)
